from unittest.mock import patch

from django.db.models import prefetch_related_objects
from django.utils import timezone

from arkindex.ponos.models import FINAL_STATES, State
from arkindex.process.models import ProcessMode
from arkindex.project.tests import FixtureAPITestCase


class TestModels(FixtureAPITestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.process = cls.corpus.processes.create(
            creator=cls.user,
            mode=ProcessMode.Workers,
        )
        cls.process.run()
        cls.task1 = cls.process.tasks.first()
        cls.nonce = b"42" + b"0" * 14

    def test_is_final(self):
        for state in State:
            self.task1.state = state
            self.task1.save()
            if state in FINAL_STATES:
                self.assertTrue(
                    self.task1.is_final(), msg="{} should be final".format(state)
                )
                self.assertTrue(
                    self.process.is_final, msg="{} should be final".format(state)
                )
            else:
                self.assertFalse(
                    self.task1.is_final(), msg="{} should not be final".format(state)
                )
                self.assertFalse(
                    self.process.is_final, msg="{} should not be final".format(state)
                )

    def test_requires_gpu(self):
        """
        Check the GPU generated requirements rules
        """
        # Default task does not need GPU
        self.assertFalse(self.task1.requires_gpu)

    @patch("arkindex.ponos.models.timezone")
    def test_task_expiry_default(self, timezone_mock):
        timezone_mock.now.return_value = timezone.datetime(3000, 1, 12).astimezone()
        # Expecting a default expiry 30 days after timezone.now
        expected_expiry = timezone.datetime(3000, 2, 11).astimezone()

        # A process with no tasks has no expiry
        self.process.tasks.all().delete()
        self.assertIsNone(self.process.expiry)

        task1 = self.process.tasks.create(
            command="do something --like this",
            slug="test_task_1",
            run=1,
            depth=0
        )
        task2 = self.process.tasks.create(
            command="do something else --like that",
            slug="test_task_2",
            run=1,
            depth=0
        )

        self.assertEqual(task1.expiry, expected_expiry)
        self.assertEqual(task2.expiry, expected_expiry)
        self.assertEqual(self.process.expiry, expected_expiry)

        # Override a task's expiry
        custom_expiry = timezone.datetime(3000, 4, 20).astimezone()
        task2.expiry = custom_expiry
        task2.save()

        # The workflow's expiry should be the latest expiry
        self.assertEqual(self.process.expiry, custom_expiry)

    def test_process_expiry_query_count(self):
        """
        Workflow.expiry causes an SQL query only when tasks are not prefetched
        """
        with self.assertNumQueries(1):
            self.assertEqual(self.process.expiry, self.task1.expiry)

        # Request the expiry again: it is not cached, there still is an SQL query
        with self.assertNumQueries(1):
            self.assertEqual(self.process.expiry, self.task1.expiry)

        prefetch_related_objects([self.process], "tasks")
        with self.assertNumQueries(0):
            self.assertEqual(self.process.expiry, self.task1.expiry)

    def test_process_get_state(self):
        with self.assertNumQueries(1):
            self.assertEqual(self.process.state, State.Unscheduled)

        with self.assertNumQueries(1):
            self.assertEqual(self.process.get_state(0), State.Unscheduled)

        with self.assertNumQueries(1):
            self.assertEqual(self.process.get_state(1), State.Unscheduled)

        # Negative run numbers should not result in any SQL query, since we know they are always empty
        with self.assertNumQueries(0):
            self.assertEqual(self.process.get_state(-1), State.Unscheduled)

        self.task1.state = State.Running
        self.task1.save()

        with self.assertNumQueries(1):
            self.assertEqual(self.process.state, State.Running)

        with self.assertNumQueries(1):
            self.assertEqual(self.process.get_state(0), State.Running)

        with self.assertNumQueries(1):
            self.assertEqual(self.process.get_state(1), State.Unscheduled)

        with self.assertNumQueries(0):
            self.assertEqual(self.process.get_state(-1), State.Unscheduled)

        prefetch_related_objects([self.process], "tasks")

        with self.assertNumQueries(0):
            self.assertEqual(self.process.state, State.Running)
            self.assertEqual(self.process.get_state(0), State.Running)
            self.assertEqual(self.process.get_state(1), State.Unscheduled)
            self.assertEqual(self.process.get_state(-1), State.Unscheduled)
