from django.conf import settings
from django.contrib.auth.password_validation import validate_password
from django.contrib.auth.tokens import default_token_generator
from django.core.exceptions import ValidationError
from django.utils.http import urlsafe_base64_decode
from drf_spectacular.utils import extend_schema_field, inline_serializer
from rest_framework import serializers

from arkindex.users.models import User


def validate_user_password(user, data):
    password = data.get("password")
    if password:
        try:
            validate_password(password, user)
        except ValidationError as e:
            raise serializers.ValidationError({"password": e.messages})
    return data


class SimpleUserSerializer(serializers.ModelSerializer):

    class Meta:
        model = User
        fields = ("id", "email", "display_name")


class UserSerializer(SimpleUserSerializer):

    features = serializers.SerializerMethodField(read_only=True)

    class Meta(SimpleUserSerializer.Meta):
        fields = SimpleUserSerializer.Meta.fields + (
            "password",
            "verified_email",
            "is_admin",
            "auth_token",
            "features",
        )
        extra_kwargs = {
            "id": {"read_only": True},
            "email": {"read_only": True},
            "password": {
                "write_only": True,
                "style": {
                    # Use a password input in DRF's HTML templates
                    "input_type": "password",
                },
            },
            "verified_email": {"read_only": True},
            "is_admin": {"read_only": True},
            "auth_token": {"read_only": True},
        }

    @extend_schema_field(inline_serializer(
        name="Features",
        fields={
            name: serializers.BooleanField()
            for name in settings.ARKINDEX_FEATURES.keys()
        }
    ))
    def get_features(self, *args, **kwargs):
        return settings.ARKINDEX_FEATURES

    def update(self, instance, validated_data):
        if "password" in validated_data:
            instance.set_password(validated_data.pop("password"))
        return super().update(instance, validated_data)


class NewUserSerializer(serializers.ModelSerializer):
    """
    A serializer that allows non authenticated users to create an account with at
    least an email address
    """
    password = serializers.CharField(
        required=False,
        write_only=True,
        style={"input_type": "password"}
    )

    class Meta:
        model = User
        fields = (
            "display_name",
            "email",
            "password"
        )

    def validate(self, data):
        data = super().validate(data)
        # Instantiate user to validate password
        user = User(**data)
        return validate_user_password(user, data)

    def create(self, validated_data):
        return User.objects.create_user(**validated_data)


class EmailLoginSerializer(serializers.Serializer):

    email = serializers.EmailField()
    password = serializers.CharField(write_only=True, style={"input_type": "password"})


class PasswordResetSerializer(serializers.Serializer):
    """
    A serializer that only takes an e-mail as the input.
    This could use a ModelSerializer on User; however, just as the Django
    default password reset view does, there should not be any errors when
    asking for a password reset on a wrong e-mail, because that would give
    account existence information to attackers.
    """
    email = serializers.EmailField()


class PasswordResetConfirmSerializer(serializers.Serializer):
    """
    Second step of a password reset: when the user clicked the link in the email.
    The frontend will send the base64-encoded user ID, reset token and new password.
    """
    uidb64 = serializers.CharField(write_only=True)
    token = serializers.CharField(write_only=True)
    password = serializers.CharField(write_only=True)

    def validate(self, data):
        # Never fail, even with an invalid UUID or non-existent user
        try:
            pk = urlsafe_base64_decode(data["uidb64"]).decode()
            user = User.objects.get(pk=pk)
        except (TypeError, ValueError, User.DoesNotExist):
            user = None

        valid_token = user is not None and default_token_generator.check_token(user, data["token"])

        validate_user_password(user, data)

        return {"user": user, "valid_token": valid_token, "password": data["password"]}

    def save(self):
        user = self.validated_data["user"]
        if not user or not self.validated_data["valid_token"]:
            return
        user.set_password(self.validated_data["password"])
        user.save()


class JobSerializer(serializers.Serializer):
    """
    Serializers a RQ job.
    """
    id = serializers.CharField(read_only=True)
    description = serializers.CharField(read_only=True)
    progress = serializers.FloatField(min_value=0, max_value=1, read_only=True, allow_null=True)
    status = serializers.SerializerMethodField()
    enqueued_at = serializers.DateTimeField(read_only=True, allow_null=True)
    started_at = serializers.DateTimeField(read_only=True, allow_null=True)
    ended_at = serializers.DateTimeField(read_only=True, allow_null=True)

    def get_status(self, instance) -> str:
        """
        Avoid causing more Redis queries to fetch a job's current status
        Note that a job status is part of a JobStatus enum,
        but the enum is just a plain object and not an Enum for Py2 compatibility.
        """
        return instance.get_status(refresh=False)
